from .hive_operations import HiveOperations
from .kafka_operations import KafkaOperations
from .elastic_operations import ElasticOperations
from .http_request_operations import HTTPRequestOperations
from .slack_operations import SlackOperations
from .bigquery_operations import BigQueryOperations, execute_from_bq_file
from .logging_operations import LoggingOperations
from .bucket_operations import BucketOperations

from .read_from_bq_file import bq_command_parser



"""_summary_

Returns:
    _type_: _description_
"""

import re


def bq_command_parser(bq_file):
    """Function to Read BigQuery .bq files and convert them into a list of BigQuery Commands.
    Args:
        bq_file (_type_): Complete Path of the BigQuery File

    Returns:
        _type_: list of BigQuery Commands
    """
    with open(bq_file, "r", encoding="utf-8") as file:
        text = file.readlines()

    # First concat all list elements to get one string
    text = "".join(text)

    # Secondly, remove block comment
    text = re.sub(r"/\*.*?\*/", "", text, flags=re.DOTALL)

    # Thirdly, remove line comments that are we are sure with
    # Note that comments that contain ' or " may remains after this filter
    text = re.sub(r"^\-\-.*", "\n", text, flags=re.M)
    text = re.sub(r"\-\-(?!(.*\'|.*\")).*\n", "\n", text, flags=re.M)

    previous_text = ""
    while previous_text != text:
        previous_text = text
        
        # Find string that are specified with ' or " and contains --
        # Those strings' locations are stored in string_spans list
        # From string_spans find dangerous ones and stored dangerous_spans
        string_spans = []
        dangerous_spans = []
        for find in re.finditer("(('|\").*?[\-]+.*?('|\"))", text, flags=re.M):
            control_start = string_spans[-1][-1] if string_spans != [] else 0
            if not re.search(r'\-\-', text[control_start:find.start()]) or \
                re.search(r'\n', text[control_start:find.start()]):
                dangerous_spans.append((find.start(), find.end()))
            string_spans.append((find.start(), find.end()))
        
        # Now find comments contains string specification like in the following:
        # --comment_1 'string
        # And check that those findings are located in dangerous spans
        # if not add to slice spans
        slice_spans = []
        for find in re.finditer(r"\-{2,}(?!.*\-\-).*", text, flags=re.M):
            safe = True
            for span in dangerous_spans:
                if find.start() >= span[0] and find.start() <= span[1]:
                    safe = False
                    break
            if safe:
                slice_spans.append((find.start(), find.end()))

        # By looking reverse of slice spans remove commented findings
        # The reason reverse list is using is if you use it ordered the location would be changed
        for start, end in reversed(slice_spans):
            text = text[:start] + text[end:]

    text = re.sub(r"(\n|\s|\t)+", " ", text)
    # add ; to last query if it doesnt have it
    if text[0] != ";":
        text = text + ";"
    # return [query.strip() for query in text.split(';')][:-1]
    return [query.strip() for query in text.split(";") if query.strip() != ""]


""" This Module is created to enable Hepsiburada Data Science to communicate with Hive """

import os
import time
from random import randint

from colorama import Fore, init ##Style
from pyhive import hive
import pandas as pd
import numpy as np
from paramiko import SSHClient, AutoAddPolicy
from scp import SCPClient

init(autoreset=True)

class HiveOperations:
    """ This class is created to enable Hepsiburada Data Science to communicate with Hive """
    _implemented_returns = ['dataframe', 'numpy', 'list', 'dict']

    def __init__(self, **kwargs):
        """ Initialize the class
        :param HIVE_HOST: Hive Server 2 host
        :param HIVE_PORT: Hive Server 2 port
        :param APP_USER_USERNAME: Hive Server 2 username
        :param APP_USER_PASS: Hive Server 2 password
        """
        self.hive_host = kwargs.get("HIVE_HOST")
        self.hive_port = kwargs.get("HIVE_PORT")
        self.hive_username = kwargs.get("HIVE_USER")
        self.hive_password = kwargs.get("HIVE_PASS")
        self.hadoop_edge_host = kwargs.get("HADOOP_EDGE_HOST")
        self.hive_connection = None

    def __repr__(self):
        return f'HiveOperations(HIVE_HOST={self.hive_host}, HIVE_PORT={self.hive_port}, ' \
               f'HIVE_USER={self.hive_username}, HIVE_PASS="XXXXXXXX")'

    def connect_to_hive(self):
        """
        This function is used to connect to Hive Server 2 and return the connection object.
        :return: Hive Server 2 connection object
        """
        try:
            self.hive_connection = hive.connect(
                                    host=self.hive_host,
                                    port=self.hive_port,
                                    username=self.hive_username,
                                    password=self.hive_password,
                                    auth='LDAP')
        except TypeError:
            raise TypeError(Fore.RED + "Failed to connect to Hive Server 2: "
                            "VPN is possibly not connected") from None

        except ValueError:
            raise ValueError(Fore.RED + "Failed to connect to Hive Server 2: "
                            "Credentials are incorrect or missing ") from None
        print(Fore.GREEN + "Connection Succeded !!")

    def disconnect_from_hive(self):
        """
        This function is used to disconnect from Hive Server 2.
        :param hive_connection: Hive Server 2 connection object
        :return: None
        """
        self.hive_connection.close()
        print(Fore.GREEN + "Connection Closed! ")

    def execute_query(self, query, return_type='dataframe', **kwargs):
        """
        This function is used to execute the query and return data in different types.
        :param hive_connection: Hive Server 2 connection object
        :param query: Hive query
        :param return_type: Type of return data
        :param kwargs return_colums: Return Columns for numpy array and list types.
        :return: Data frame
        """
        if return_type not in HiveOperations._implemented_returns:
            raise NotImplementedError(Fore.RED + f'Return type {return_type} not implemented !!')

        cursor = self.hive_connection.cursor()
        execution_start = time.time()
        cursor.execute(query)
        execution_duration = time.time() - execution_start
        print(Fore.YELLOW + f'Query executed in {execution_duration:.2f} seconds !')
        columns = HiveOperations._rename_columns(cursor)

        if return_type == 'dataframe':
            data = pd.DataFrame(cursor.fetchall(), columns=columns)
        elif return_type == 'numpy':
            data = np.array(cursor.fetchall())
        elif return_type == 'list':
            data = cursor.fetchall()
        elif return_type == 'dict':
            data = [dict(zip(columns, row)) for row in cursor.fetchall()]

        cursor.close()
        if kwargs.get('return_columns'):
            return data, columns
        return data, None

    def create_insert_table(self, query) -> None:
        """
        This function is used to create the table and insert data into the table.
        :param query: Hive query
        :param hive_connection: Hive Server 2 connection object
        :return: None
        """
        cursor = self.hive_connection.cursor()
        cursor.execute(query)
        cursor.close()

    def send_files_to_hive(self, target_table, source_data, columns=None):
        """
        This function is used to send files to Hive Server 2.
        :param target_table: Target Hive table
        :param source_data: Source data
        :param columns: Columns of the source data
        :return: None
        """
        unique_filename = str(randint(0, 100000))
        transfer_start = time.time()
        ssh_client = self._set_ssh_client()

        csv_created= False
        edge_node_path = f"/home/{self.hive_username}/Python_CSV"
        hdfs_path = f"/user/{self.hive_username}/projects/Python_CSV"

        # Check if Edge Node & HDFS Path exist
        edge_node_dir_check_cmd = f"ls -d {edge_node_path}"
        hdfs_dir_check_cmd = f"hdfs dfs -test -d {hdfs_path}"

        _, stdout, _ = ssh_client.exec_command(edge_node_dir_check_cmd)
        if stdout.channel.recv_exit_status() != 0:
            print(Fore.RED + f"Edge Node Path {edge_node_path} does not exist")
            print(Fore.YELLOW + f"Creating Edge Node Path {edge_node_path}")
            ssh_client.exec_command(f"mkdir {edge_node_path}")

        _, stdout, _ = ssh_client.exec_command(hdfs_dir_check_cmd)
        if stdout.channel.recv_exit_status() != 0:
            print(Fore.RED + f"HDFS Path {hdfs_path} does not exist")
            print(Fore.YELLOW + f"Creating HDFS Path {hdfs_path}")
            ssh_client.exec_command(f"hdfs dfs -mkdir {hdfs_path}")

        try:
            # Creating a csv file
            if str(type(source_data)) ==  "<class 'numpy.ndarray'>":
                np.savetxt(unique_filename + ".csv", source_data,
                                    delimiter="\t",header='\t'.join(columns),comments='')
                source_data = unique_filename + ".csv"
                csv_created = True
                # print("Temporary csv file '" + unique_filename +".csv' created.")

            elif str(type(source_data)) ==  "<class 'pandas.core.frame.DataFrame'>":
                source_data.to_csv(unique_filename + ".csv",index=False,sep="\t")
                source_data = unique_filename + ".csv"
                csv_created = True
                # print("Temporary csv file '" + unique_filename +".csv' created.")


            _, stdout, _ = ssh_client.exec_command(f"mkdir {edge_node_path}/{unique_filename};")
            if stdout.channel.recv_exit_status() != 0:
                raise RuntimeError(Fore.RED + "Failed to create directory on edge node")

            with SCPClient(ssh_client.get_transport()) as scp:
                # uploads target CSV to edgenode
                scp.put(source_data, recursive=True, remote_path=f"{edge_node_path}/{unique_filename}/")
            print(Fore.GREEN + "Files sent to EdgeNode succesfully!")

            edge_node_to_hdfs_command_list = [
                f"hadoop fs -rm -r {hdfs_path}/{unique_filename};",
                f"hadoop fs -mkdir {hdfs_path}/{unique_filename}/;",
                f"hadoop fs -put {edge_node_path}/{unique_filename}/*.csv {hdfs_path}/{unique_filename};",
                f"rm -r {edge_node_path}/{unique_filename};"
                ]

            for command in edge_node_to_hdfs_command_list:
                _, stdout, _ = ssh_client.exec_command(command)
                if stdout.channel.recv_exit_status() == 0:
                    print(Fore.GREEN + f"Command executed: {command}")
                else:
                    print(Fore.RED + f"Command failed: {command}")

            print(Fore.GREEN + "EdgeNode - HDFS transfer completed! \nEdgeNode is clean again!")

            # reading csv file's column types and make column types suitable for hql
            dataframe = pd.read_csv(source_data, sep='\t')
            column_name = dataframe.columns
            dtype_dict = {
                "object": "STRING",
                "int64": "INT",
                "float64": "FLOAT",
                "bool": "BOOLEAN"
                }
            column_type = [dtype_dict[str(item)] for item in dataframe.dtypes]
            columns = []

            for item in zip(column_name, column_type):
                columns.append(" ".join(item))
            column_details = ",".join(columns)

            # numpy,pandas csvleri siliniyor
            if csv_created:
                os.remove(source_data)
                csv_created = False

            # creating external table
            print(Fore.GREEN + "Creating external table...")
            self.create_insert_table(f"DROP TABLE IF EXISTS {target_table}_ext")
            self.create_insert_table(f"CREATE EXTERNAL TABLE {target_table}_ext"
                                    f"({column_details}) "
                                    f'ROW FORMAT DELIMITED FIELDS TERMINATED BY "\t" '
                                    f'LOCATION "{hdfs_path}/{unique_filename}/" '
                                    f'TBLPROPERTIES("skip.header.line.count"="1")')

            self.create_insert_table(f"DROP TABLE IF EXISTS {target_table}")
            self.create_insert_table("SET hive.auto.convert.join=false")
            self.create_insert_table("SET hive.tez.container.size=16384")
            self.create_insert_table("SET hive.tez.java.opts=-Xmx13108m")
            self.create_insert_table(f"CREATE TABLE {target_table} AS "
                                    f"SELECT * FROM {target_table}_ext")
            self.create_insert_table(f"DROP TABLE IF EXISTS {target_table}_ext")

            print(Fore.GREEN + "Table creation Success!")
            transfer_end = time.time()
            print(Fore.YELLOW +
                    f"Table Creation took: {transfer_end - transfer_start:.2f} seconds!")

            ssh_client.exec_command(f"hadoop fs -rm -r {hdfs_path}/{unique_filename};")
            print(Fore.GREEN + "HDFS is clean again!")

        except Exception as exc:
            ssh_client.exec_command(f"hadoop fs -rm -r {hdfs_path}/{unique_filename};")
            print(Fore.GREEN + "HDFS is clean!")
            ssh_client.exec_command(f"rm -r {edge_node_path}/{unique_filename};")
            print(Fore.GREEN + "EdgeNode is clean!")
            raise exc

    def _set_ssh_client(self):
        ssh_client = SSHClient()
        ssh_client.load_system_host_keys()
        ssh_client.set_missing_host_key_policy(AutoAddPolicy())
        ssh_client.connect(hostname=self.hadoop_edge_host,
                            username=self.hive_username,
                            password=self.hive_password, look_for_keys=False)
        return ssh_client

    @staticmethod
    def _rename_columns(cursor) -> list:
        """
        This function takes cursor description and rename the columns of the data frame.
        :param column_names: Column names
        :return: List of column names
        """
        columns = [column[0].split('.')[-1] for column in cursor.description]
        return columns

    @staticmethod
    def _parse_sql_query(query):
        """
        This function is used to parse the sql query and return the query with the table name.
        :param query: Hive query
        :return: Hive query with table name
        """

        ## TO BE WRITTEN !!

        query_list = query.split(' ')
        table_name = query_list[2]
        return table_name

    @staticmethod
    def _execute_sql_from_file(query_file, hive_connection):
        """
        This function is used to execute the sql query from file and return the data frame.
        :param query_file: Hive query file
        :param hive_connection: Hive Server 2 connection object
        :return: Data frame
        """

        ## TO BE WRITTEN !!

        with open(query_file, 'r', encoding='utf-8') as hiveql_file:
            query = hiveql_file.read()
        return HiveOperations.execute_query(hive_connection, query)

    @property
    def hive_host(self):
        """Hive Server 2 host"""
        return self._hive_host

    @hive_host.setter
    def hive_host(self, provided_hive_host):
        """Hive Server 2 host setter"""
        if provided_hive_host is not None:
            self._hive_host = str(provided_hive_host)
        elif os.environ.get("HIVE_HOST"):
            self._hive_host = os.environ.get("HIVE_HOST")
        else:
            self._hive_host = None
            print(Fore.RED + "Warning!! Hive host is not specified")

    @property
    def hive_port(self):
        """Hive Server 2 port"""
        return self._hive_port

    @hive_port.setter
    def hive_port(self, provided_hive_port):
        """Hive Server 2 port setter"""
        if provided_hive_port is not None:
            self._hive_port = str(provided_hive_port)
        elif os.environ.get("HIVE_PORT"):
            self._hive_port = os.environ.get("HIVE_PORT")
        else:
            self._hive_port = None
            print(Fore.RED + "Warning!! Hive port is not specified")

    @property
    def hive_username(self):
        """Hive Server 2 username"""
        return self._hive_username

    @hive_username.setter
    def hive_username(self, provided_hive_username):
        """Hive Server 2 username setter"""
        if provided_hive_username is not None:
            self._hive_username = str(provided_hive_username)
        elif os.environ.get("HIVE_USERNAME"):
            self._hive_username = os.environ.get("HIVE_USERNAME")
        else:
            self._hive_username = None
            print(Fore.RED + "Warning!! Hive username is not specified")

    @property
    def hive_password(self):
        """Hive Server 2 password"""
        return self._hive_password

    @hive_password.setter
    def hive_password(self, provided_hive_password):
        """Hive Server 2 password setter"""
        if provided_hive_password is not None:
            self._hive_password = str(provided_hive_password)
        elif os.environ.get("HIVE_PASSWORD"):
            self._hive_password = os.environ.get("HIVE_PASSWORD")
        else:
            self._hive_password = None
            print(Fore.RED + "Warning!! Hive password is not specified")

    @property
    def hadoop_edge_host(self):
        """Hadoop Edge host"""
        return self._hadoop_edge_host

    @hadoop_edge_host.setter
    def hadoop_edge_host(self, provided_hadoop_edge_host):
        """Hadoop Edge host setter"""
        if provided_hadoop_edge_host is not None:
            self._hadoop_edge_host = str(provided_hadoop_edge_host)
        elif os.environ.get("HADOOP_EDGE_HOST"):
            self._hadoop_edge_host = os.environ.get("HADOOP_EDGE_HOST")
        else:
            self._hadoop_edge_host = None
            print(Fore.RED + "Warning!! Hadoop Edge Host is not specified")



""" This Module is created to enable Hepsiburada Data Science team to log Python Applications. """

import os
import logging

# from logging.handlers import RotatingFileHandler
# from pathlib import Path


from colorama import Fore, init
from google.oauth2 import service_account
import google.cloud.logging as gcp_logging
from google.cloud.logging_v2.handlers import CloudLoggingHandler

init(autoreset=True)


class LoggingOperations:
    """This class is created to enable Hepsiburada Data Science team to log Python Applications."""

    def __init__(self, **kwargs) -> None:
        self.gcp_key = kwargs.get("gcp_key_path")
        self.logger_name = kwargs.get("logger_name")
        self.project = kwargs.get("project")
        self.submodule = kwargs.get("submodule")
        self.credentials = kwargs.get("gcp_key_path")
        self.logger = None

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(Project ID: {self.credentials.project_id}, "
            f"Service Account: {self.credentials._service_account_email.split('@')[0]}, "
            f"Logger Name: {self.logger_name}, "
            f"Project: {self.project}, Submodule: {self.submodule})"
        )

    def set_logger(self):
        """This function is to set logger."""
        # root_dir = Path(__file__).parent.parent
        # log_file = f"reports/logs/{self.project}_{self.submodule}.log"
        # log_file_path = root_dir.joinpath(log_file)
        self.logger = logging.getLogger(self.logger_name)
        self.logger.setLevel(logging.DEBUG)

        # file_handler = RotatingFileHandler(
        #     log_file_path, "a", 10 * 1024 * 1024, 5, "utf-8"
        # )
        # file_handler.setLevel(logging.DEBUG)
        formatter = logging.Formatter(
            "%(asctime)s - %(name)s - %(levelname)s - %(funcName)2s() - %(message)s"
        )

        gcloud_logging_client = gcp_logging.Client(
            project=self.credentials.project_id, credentials=self.credentials
        )
        gcloud_logging_handler = CloudLoggingHandler(
            gcloud_logging_client, name=self.logger_name
        )
        # file_handler.setFormatter(formatter)
        gcloud_logging_handler.setFormatter(formatter)
        gcloud_logging_handler.setLevel(logging.DEBUG)

        # self.logger.addHandler(file_handler)
        self.logger.addHandler(gcloud_logging_handler)
        self.logger.info(msg="Logger has been initiated! ")

    @property
    def gcp_key(self):
        """This function is to get GCP key."""
        return self._gcp_key

    @gcp_key.setter
    def gcp_key(self, provided_gcp_key):
        """This function is to set GCP key."""
        if provided_gcp_key is not None:
            self._gcp_key = str(provided_gcp_key)
            self.credentials = str(provided_gcp_key)
        elif os.environ.get("SERVICE_ACCOUNT_KEY_PATH"):
            self._gcp_key = os.environ.get("SERVICE_ACCOUNT_KEY_PATH")
        else:
            self._gcp_key = None
            print(
                Fore.RED + "Warning!! GCP Key Path for Service Account is not specified"
            )

    @property
    def credentials(self):
        """This function is to get credentials."""
        return self._credentials

    @credentials.setter
    def credentials(self, provided_credentials):
        """This function is to set credentials."""
        if provided_credentials is not None:
            self._credentials = service_account.Credentials.from_service_account_file(
                self.gcp_key,
                scopes=["https://www.googleapis.com/auth/cloud-platform"],
            )
        elif os.environ.get("SERVICE_ACCOUNT_KEY_PATH"):
            self._credentials = service_account.Credentials.from_service_account_file(
                self.gcp_key,
                scopes=["https://www.googleapis.com/auth/cloud-platform"],
            )
        else:
            self._credentials = None
            print(
                Fore.RED + "Warning!! Credentials for Service Account is not specified"
            )



""" This Module is created to enable Hepsiburada Data Science to communicate with BigQuery. """

import os
import sys
import time
import warnings
import numpy as np

from colorama import Fore, init  ##Style
from google.cloud import bigquery
from google.oauth2 import service_account
from google.api_core.exceptions import NotFound, BadRequest

from .read_from_bq_file import bq_command_parser

init(autoreset=True)


class BigQueryOperations(bigquery.Client):
    """This class is created to enable Hepsiburada Data Science to communicate with BigQuery"""

    _implemented_returns = ["dataframe", "array", "arrow", None]

    def __init__(self, **kwargs) -> None:
        self.gcp_key = kwargs.get("gcp_key_path")
        self.credentials = kwargs.get("gcp_key_path")
        super().__init__(
            project=self.credentials.project_id,
            credentials=self.credentials,
            location="EU",
        )

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(Project ID: {self.credentials.project_id}, "
            f"Service Account: {self.credentials._service_account_email.split('@')[0]})"
        )

    def get_bq_table(self, table_name):
        """This function is to get BQ table."""
        return self.get_table(table_name)

    def execute_query(self, query_string, return_type="dataframe", **kwargs):
        """This function is to query BQ."""
        if return_type not in BigQueryOperations._implemented_returns:
            raise NotImplementedError(
                Fore.RED + f"Return type {return_type} not implemented !!"
            )

        # Verbosity based on the environment !
        env = kwargs.get("environ")
        verbose = True
        if env in ["dev", "qa"]:
            kwargs.pop("environ")
        elif env == "prod":
            verbose = False
            kwargs.pop("environ")

        job_config = bigquery.QueryJobConfig()
        job_config.dry_run = True
        try:
            job_result = self.query(query_string, job_config=job_config, **kwargs)
        except BadRequest as bad_request_message:
            raise RuntimeError(
                Fore.RED + f"Query failed with error: {bad_request_message}"
            ) from bad_request_message

        if job_result.total_bytes_processed / (1024 * 1024 * 1024) > 100:
            if verbose:
                print(
                    Fore.RED
                    + "Query is trying to process "
                    + f"{(job_result.total_bytes_processed)/(1024*1024*1024):.2f} GB of data!"
                )

        job_config.dry_run = False
        execution_start = time.time()
        job_result = self.query(query_string, job_config=job_config, **kwargs)
        query_result = job_result.result()
        execution_duration = time.time() - execution_start

        # To check if create table query is executed successfully.
        # To prevent empty table creation.
        if job_result.statement_type == "CREATE_TABLE_AS_SELECT":
            destination_table_num_rows = self.get_table(job_result.destination).num_rows
            if destination_table_num_rows == 0:
                raise ValueError(
                    Fore.RED + f"Table {job_result.destination} is empty !!"
                )

        if verbose:
            print(Fore.YELLOW + f"Query executed in {execution_duration:.2f} seconds !")
            print(
                Fore.YELLOW
                + "Query processed "
                + f"{(job_result.total_bytes_processed)/(1024*1024):.2f} MB !"
            )

        if return_type == "dataframe":
            return query_result.to_dataframe(progress_bar_type="tqdm")
        elif return_type == "array":
            query_result=query_result.to_arrow(progress_bar_type="tqdm")
            return np.asarray(query_result, dtype=object)
        elif return_type == "arrow":
            return query_result.to_arrow(progress_bar_type="tqdm")
        elif return_type is None:
            return None

    def create_new_dataset(self, dataset_name):
        """This function is to create dataset."""
        _ = self.create_dataset(dataset_name)

    def delete_existing_dataset(self, dataset_name):
        """This function is to delete dataset.
        Note: This function will not delete the dataset if there are tables in it."""
        self.delete_dataset(dataset_name)

    def delete_existing_table(self, dataset, table):
        """This function is to delete table."""
        self.delete_table(f"{self.project}.{dataset}.{table}")

    def create_new_table(self, dataset, table_name, schema, **kwargs):
        """This function is to create table."""
        provided_schema = [
            bigquery.SchemaField(
                field["field_name"], field["field_type"], field["field_mode"]
            )
            for field in schema
        ]

        table_reference = bigquery.Table(
            f"{self.project}.{dataset}.{table_name}", schema=provided_schema
        )
        self.create_table(table_reference, provided_schema, **kwargs)

    def load_data_to_table(self, dataset, table_name, data_frame, **kwargs):
        """This function is to create a table and load data from dataframe into it."""
        if dataset in [dataset.dataset_id for dataset in self.list_datasets()]:
            job_config = bigquery.job.LoadJobConfig()
            if kwargs.get("overwrite"):
                job_config.write_disposition = bigquery.WriteDisposition.WRITE_TRUNCATE
                # overwrite asıl class'ta yok o yüzden asıl class'a paslamamak için kaldırıldı.
                # sonraki sürümlerde overwrite parametreye alınınca buna gerek kalmayacak.
                kwargs.pop("overwrite")
            else:
                job_config.write_disposition = bigquery.WriteDisposition.WRITE_APPEND
                if kwargs.get("overwrite") is None:
                    warnings.showwarning(
                        message="""Overwrite is not provided. Default is False. """
                        """Please provide overwrite parameter since it will be required in future releases.""",
                        category=DeprecationWarning,
                        filename=__name__ + ".py",
                        lineno=4,
                        file=sys.stdout,
                    )
                # overwrite asıl class'ta yok o yüzden asıl class'a paslamamak için kaldırıldı.
                # sonraki sürümlerde overwrite parametreye alınınca buna gerek kalmayacak.
                else:
                    kwargs.pop("overwrite")
            job = self.load_table_from_dataframe(
                data_frame,
                f"{self.project}.{dataset}.{table_name}",
                job_config=job_config,
                **kwargs,
            )
            job.result()
        else:
            raise ValueError(Fore.RED + f"Dataset {dataset} does not exist !")

    def insert_rows_into_existing_table(self, dataset, table, data):
        """This function is to insert rows into existing table."""
        table_reference = self.get_bq_table(f"{self.project}.{dataset}.{table}")
        try:
            insert_result = self.insert_rows_from_dataframe(table_reference, data)
        except NotFound:
            insert_result = self.insert_rows_from_dataframe(table_reference, data)
        if len(insert_result[0]) > 0:
            print(Fore.RED + f"Insertion failed for {insert_result[0]} !")

    @property
    def gcp_key(self):
        """This function is to get GCP key."""
        return self._gcp_key

    @gcp_key.setter
    def gcp_key(self, provided_gcp_key):
        """This function is to set GCP key."""
        if provided_gcp_key is not None:
            self._gcp_key = str(provided_gcp_key)
            self.credentials = str(provided_gcp_key)
        elif os.environ.get("SERVICE_ACCOUNT_KEY_PATH"):
            self._gcp_key = os.environ.get("SERVICE_ACCOUNT_KEY_PATH")
        else:
            self._gcp_key = None
            print(
                Fore.RED + "Warning!! GCP Key Path for Service Account is not specified"
            )

    @property
    def credentials(self):
        """This function is to get credentials."""
        return self._credentials

    @credentials.setter
    def credentials(self, provided_credentials):
        """This function is to set credentials."""
        if provided_credentials is not None:
            self._credentials = service_account.Credentials.from_service_account_file(
                self.gcp_key,
                scopes=["https://www.googleapis.com/auth/cloud-platform"],
            )
        elif os.environ.get("SERVICE_ACCOUNT_KEY_PATH"):
            self._credentials = service_account.Credentials.from_service_account_file(
                self.gcp_key,
                scopes=["https://www.googleapis.com/auth/cloud-platform"],
            )
        else:
            self._credentials = None
            print(
                Fore.RED + "Warning!! Credentials for Service Account is not specified"
            )


def execute_from_bq_file(bq_client, bq_file_path, **kwargs):
    """This function is to execute bq file in designated environment."""
    bq_commands = bq_command_parser(bq_file_path)
    transformed_bq_commands = []

    for bq_command in bq_commands:
        if kwargs.get("config") is not None:
            environ = kwargs.get("config")["env_identifier"]
            target_db_prefix = kwargs.get("config")["db_prefix"]
            if environ == "qa":
                source_db_prefix = target_db_prefix.replace("qa", "dev")
                bq_command = bq_command.replace(source_db_prefix, target_db_prefix)
            elif environ == "prod":
                source_db_prefix = target_db_prefix.replace("prod", "dev")
                bq_command = bq_command.replace(source_db_prefix, target_db_prefix)
        else:
            environ = "dev"
            print(Fore.YELLOW + "Default Env is DEV")

        if kwargs.get("verbose"):
            print(Fore.GREEN + f"Executing: {bq_command}")
        if not kwargs.get("dependent_queries"):
            bq_client.execute_query(bq_command, return_type=None, environ=environ)
        else:
            transformed_bq_commands.append(bq_command)
    if kwargs.get("dependent_queries"):
        multi_query = ";".join(transformed_bq_commands)
        bq_client.execute_query(multi_query, return_type=None, environ=environ)



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['heps_ds_utils']

package_data = \
{'': ['*']}

install_requires = \
['PyHive>=0.6.5,<0.7.0',
 'colorama>=0.4.4,<0.5.0',
 'google-cloud-bigquery[bqstorage,pandas]>=3.0.1,<4.0.0',
 'google-cloud-logging>=3.0.0,<4.0.0',
 'google-cloud-storage>=2.5.0,<3.0.0',
 'pandas>=1.4.1,<2.0.0',
 'paramiko>=2.10.3,<3.0.0',
 'protobuf==3.20.0',
 'scp>=0.14.4,<0.15.0',
 'thrift-sasl>=0.4.3,<0.5.0',
 'thrift>=0.15.0,<0.16.0',
 'tqdm>=4.64.0,<5.0.0']

extras_require = \
{':sys_platform == "linux" or sys_platform == "darwin"': ['sasl>=0.3.1,<0.4.0']}

setup_kwargs = {
    'name': 'heps-ds-utils',
    'version': '0.4.6a9',
    'description': 'A Module to enable Hepsiburada Data Science Team to utilize different tools.',
    'long_description': '# Hepsiburada Data Science Utilities\n\nThis module includes utilities for Hepsiburada Data Science Team.\n\n- Library is available via PyPi. \n- Library can be downloaded using pip as follows: `pip install heps-ds-utils`\n- Existing library can be upgraded using pip as follows: `pip install heps-ds-utils --upgrade`\n\n***\n## Available Modules\n\n1. Hive Operations\n\n```python\nfrom heps_ds_utils import HiveOperations\n\n# A connection is needed to be generated in a specific runtime.\n# There are 3 ways to set credentials for connection.\n\n# 1) Instance try to set default credentials from Environment Variables.\nhive_ds = HiveOperations()\nhive_ds.connect_to_hive()\n\n# 2) One can pass credentials to instance initiation to override default.\nhive_ds = HiveOperations(HIVE_HOST="XXX", HIVE_PORT="YYY", HIVE_USER="ZZZ", HIVE_PASS="WWW", HADOOP_EDGE_HOST="QQQ")\nhive_ds = HiveOperations(HIVE_USER="ZZZ", HIVE_PASS="WWW")\nhive_ds.connect_to_hive()\n\n# 3) One can change any of the credentials after initiation using appropriate attribute.\nhive_ds = HiveOperations()\nhive_ds.hive_username = \'XXX\'\nhive_ds.hive_password = \'YYY\'\nhive_ds.connect_to_hive()\n\n#\xa0Execute an SQL query to retrieve data.\n# Currently Implemented Types: DataFrame, Numpy Array, Dictionary, List.\nSQL_QUERY = "SELECT * FROM {db}.{table}"\ndata, columns = hive_ds.execute_query(SQL_QUERY, return_type="dataframe", return_columns=False)\n\n# Execute an SQL query to create and insert data into table.\nSQL_QUERY = "INSERT INTO .."\nhive_ds.create_insert_table(SQL_QUERY)\n\n# Send Files to Hive and Create a Table with the Data.\n# Currently DataFrame or Numpy Array can be sent to Hive.\n# While sending Numpy Array columns have to be provided.\nSQL_QUERY = "INSERT INTO .."\nhive_ds.send_files_to_hive("{db}.{table}", data, columns=None)\n\n# Close the connection at the end of the runtime.\n\nhive_ds.disconnect_from_hive()\n\n```\n\n2. BigQuery Operations\n\n```python\nfrom heps_ds_utils import BigQueryOperations, execute_from_bq_file\n\n# A connection is needed to be generated in a specific runtime.\n# There are 3 ways to set credentials for connection.\n\n# 1) Instance try to set default credentials from Environment Variables.\nbq_ds = BigQueryOperations()\n\n# 2) One can pass credentials to instance initiation to override default.\nbq_ds = BigQueryOperations(gcp_key_path="/tmp/keys/ds_qos.json")\n\n# Unlike HiveOperations, initiation creates a direct connection. Absence of\n# credentials will throw an error.\n\n#\xa0Execute an SQL query to retrieve data.\n# Currently Implemented Types: DataFrame.\nQUERY_STRING = """SELECT * FROM `[project_name].[dataset_name].[table_name]` LIMIT 20"""\ndata = bq_ds.execute_query(QUERY_STRING, return_type=\'dataframe\')\n\n#\xa0Create a Dataset in BigQuery.\nbq_ds.create_new_dataset("example_dataset")\n\n# Create a Table under a Dataset in BigQuery.\nschema = [\n    {"field_name": "id", "field_type": "INTEGER", "field_mode": "REQUIRED"},\n    {"field_name": "first_name", "field_type": "STRING", "field_mode": "REQUIRED"},\n    {"field_name": "last_name", "field_type": "STRING", "field_mode": "REQUIRED"},\n    {"field_name": "email", "field_type": "STRING", "field_mode": "REQUIRED"},\n    {"field_name": "gender", "field_type": "STRING", "field_mode": "REQUIRED"},\n    {"field_name": "ip_address", "field_type": "STRING", "field_mode": "REQUIRED"}]\n\nbq_ds.create_new_table(dataset=\'example_dataset\', table_name=\'mock_data\', schema=schema)\n\n# Insert into an existing Table from Dataframe.\n# Don\'t create and insert in the same runtime.\n# Google throws an error when creation and insertion time is close.\nbq_ds.insert_rows_into_existing_table(dataset=\'example_dataset\', table=\'mock_data\', data=df)\n\n# Delete a Table.\nbq_ds.delete_existing_table(\'example_dataset\', \'mock_data\')\n\n# Delete a Dataset.\n# Trying to delete a dataset consisting of tables will throw an error.\nbq_ds.delete_existing_dataset(\'example_dataset\')\n\n# Load Dataframe As a Table. BigQuery will infer the data types.\nbq_ds.load_data_to_table(\'example_dataset\', \'mock_data\', df, overwrite=False)\n\n# To execute BQ commands sequentially from a BigQuery Script without a return statement !\nexecute_from_bq_file(bq_client=bq_ds, bq_file_path="tests/test_data/test_case_2.bq", verbose=True)\n\n```\n\n3. Logging Operations\n\n```python\nfrom heps_ds_utils import LoggingOperations\n\n# A connection is needed to be generated in a specific runtime.\n# There are 3 ways to set credentials for connection.\n\n# 1) Instance try to set default credentials from Environment Variables.\nlogger_ds = LoggingOperations()\n\n# 2) One can pass credentials to instance initiation to override default.\nlogger_ds = LoggingOperations(gcp_key_path="/tmp/keys/ds_qos.json")\n\n# Unlike HiveOperations, initiation creates a direct connection. Absence of\n# credentials will throw an error.\n\n\n```\n\n4. Bucket Operations\n\n```python\nfrom heps_ds_utils import BucketOperations\n\n# A connection is needed to be generated in a specific runtime.\n# There are 2 ways to set credentials for connection.\n\n# 1) Instance try to set default credentials from Environment Variables.\nbct_ds = BucketOperations()\n\n# 2) One can pass credentials to instance initiation to override default.\nbct_ds = BucketOperations(gcp_key_path="/tmp/keys/ds_qos.json")\n\n# Unlike HiveOperations, initiation creates a direct connection. Absence of\n# credentials will throw an error.\n\nBUCKET_NAME = "bucket-name"\n\n# Upload File using filepath.\n# Blob name is the filepath you want the file to be under the bucket.\n# Filepath is the path to the file you want to upload.\nbct_ds.upload_from_filepath(BUCKET_NAME, "project_name/dev/data/output.csv", "data/output.csv")\n\n# Upload File from memory data.\n# If you want to save the data in memory to a file, you can use this function.\nbct_ds.upload_from_memory(BUCKET_NAME, "project_name/dev/model/model.pkl", model)\n\n# Download file from bucket to filepath.\nbct_ds.download_to_filepath(BUCKET_NAME, "project_name/dev/data/sample.json", "data/sample.json")\n\n# Download data from bucket to memory.\n# If you want to save the data in a file to memory, you can use this function.\nframe = bct_ds.download_to_memory(BUCKET_NAME, "project_name/dev/data/sample.csv", "dataframe")\n\n# Delete file from bucket.\nbct_ds.delete_file_from_bucket(BUCKET_NAME, "project_name/dev/data/sample.json")\n\n# Create empty folder to bucket.\nbct_ds.create_new_folders(BUCKET_NAME, "project_name/dev/data/")\n\n```\n\nRelease Notes:\n\n0.4.4:\n- BigQueryOperations:\n    - insert_rows_into_existing_table: insertion exception handling added.\n    - insert_rows_into_existing_table: retry added. \n        - Put time between table creation and insertion.\n    - execute_query: total_bytes_processed info added.\n    - execute_query: max allowed total_bytes_processed set to 100GB.\n    - execute_query: return_type=None for Queries w/o any return.\n    - load_data_to_table: kwargs[\'overwrite\'] is added.\n        - load_data_to_table(..., overwrite=True) to overwrite to table.\n        - load_data_to_table(..., overwrite=False) to append to table.\n        - not passing overwrite kwarg will print a DeprecationWarning.\n    - execute_from_bq_file: sequential execution of BigQuery commands from\n    a file. It has its own parser. \n        - execute_from_bq_file(..., verbose=True) to print BigQuery commands to console.\n        - execute_from_bq_file(..., verbose=False) not to print BigQuery commands to console.\n\n0.4.5:\n- LoggingOperations\n    - Bug Fix in Authentication to GCP Logging !\n- BigQueryOperations\n    - Executing BQ files for different environments !\n\n0.4.6:\n- BigQueryOperations\n    - BQ Parser bug fix !\n    - BQ File Execution dependent queries\n        - Some of the queries depends on the previous command executions. For these cases:\n        dependent_queries is needed to be set to True !\n        execute_from_bq_file(\n            bq_ds,\n            "tests/test_data/test_case_4.bq",\n            verbose=True,\n            config=configs,\n            dependent_queries=True)\n    - BQ Create Table Results in Empty Table Check Added!\n        - Raises an error if CREATE TABLE ... SELECT AS ... query results in empty table.\n        - This doesn\'t work in the case of dependent_queries=True !!!\n    - 100GB limit is turned into a warning, which will not be displayed in prod env.\n    - BQ Return Types Implemented (Numpy Array and Arrow Formats)\n- LoggingOperations\n    - protobuf dependency issue resolved!\n- BucketOperations\n    - upload_from_filepath function added.\n    - upload_from_memory function added.\n    - download_to_filepath function added.\n    - download_to_memory function added.\n    - delete_file_from_bucket function added.\n    - create_new_folders function added.\n    - delete_folder function added.\n',
    'author': 'FarukBuldur',
    'author_email': 'faruk.buldur@hepsiburada.com',
    'maintainer': 'FıratÖncü',
    'maintainer_email': 'firat.oncu@hepsiburada.com',
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



""" This Module is created to enable Hepsiburada Data Science to communicate with Cloud Storage. """

import os
import json
import pickle
from io import BytesIO
import pandas as pd

from colorama import Fore, init  ##Style
from google.cloud import storage
from google.oauth2 import service_account
from google.api_core.exceptions import NotFound


init(autoreset=True)


class BucketOperations(storage.Client):
    """This class is created to enable Hepsiburada Data Science to communicate with Cloud Storage"""

    _implemented_type = ["dataframe", "json", "pickle"]

    def __init__(self, **kwargs) -> None:
        self.gcp_key = kwargs.get("gcp_key_path")
        self.credentials = kwargs.get("gcp_key_path")
        super().__init__(
            project=self.credentials.project_id,
            credentials=self.credentials
        )

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}(Project ID: {self.credentials.project_id}, "
            f"Service Account: {self.credentials._service_account_email.split('@')[0]})"
        )

    def upload_from_filepath(self,bucket_name, blob_name, filepath):
        """This function is to upload data from filepath."""
        try:
            bucket = self.bucket(bucket_name)
            blob = bucket.blob(blob_name)
            blob.upload_from_filename(filepath)

        except NotFound:
            print(f"Bucket <{bucket_name}> does not exist.")

        except Exception:
            print(f"No such file in this path: <{filepath}>")
    
    def upload_from_memory(self, bucket_name, blob_name, contents, upload_type):
        """This function is to upload data from memory."""
        if upload_type not in BucketOperations._implemented_type:
            raise NotImplementedError(
                Fore.RED + f"Download type {upload_type} not implemented !!"
            )

        try:
            bucket = self.bucket(bucket_name)
            blob = bucket.blob(blob_name)

            if upload_type == 'json':
                send_data = json.dumps(contents, indent=4).encode('utf-8')
            elif upload_type == 'pickle':
                send_data = pickle.dumps(contents)
            elif upload_type == "dataframe":
                send_data = contents.to_csv()

            blob.upload_from_string(send_data)

        except NotFound:
            print(f"Bucket <{bucket_name}> does not exist.")

    def download_to_filepath(self,bucket_name, blob_name, filepath): 
        """This function is to download data from filepath."""
        try:
            bucket = self.bucket(bucket_name)
            blob = bucket.blob(blob_name)
            if blob.exists():
                with open(filepath, 'wb') as f:
                    self.download_blob_to_file(blob, f)
            else:
                print(f"Blob <{blob_name}> does not exist.")

        except NotFound:
            print(f"Bucket <{bucket_name}> does not exist.")
        
        except Exception:
            print(f"No such file in this path: <{filepath}>")

    def download_to_memory(self,bucket_name, blob_name, download_type):
        """This function is to download data from memory."""
        if download_type not in BucketOperations._implemented_type:
            raise NotImplementedError(
                Fore.RED + f"Download type {download_type} not implemented !!"
            )

        try:
            bucket = self.bucket(bucket_name)
            blob = bucket.blob(blob_name)
            if blob.exists():
                if download_type == 'dataframe':
                    df = pd.read_csv(
                            BytesIO(
                                blob.download_as_string()
                            ) ,
                            encoding='UTF-8',
                            sep=','
                        )
                    return df

                elif download_type == 'json':
                    data = blob.download_as_string().decode('utf8').replace("'", '"')
                    load_json = json.loads(data)
                    json_data = json.dumps(load_json, indent=4, sort_keys=True)
                    return json_data

                elif download_type == 'pickle':
                    data = blob.download_as_string()
                    load_pickle = pickle.loads(data)
                    return load_pickle
            else:
                print(f"Blob <{blob_name}> does not exist.")

        except NotFound:
            print(f"Bucket <{bucket_name}> does not exist.")

    def create_new_folders(self, bucket_name, folder_name):
        """This function is to create new folder from the bucket."""
        try:
            bucket = self.bucket(bucket_name)
            dst_bucket = self.bucket(bucket_name)

            blob = bucket.blob("New_Folder_Starter/")
            new_blob = bucket.copy_blob(blob, dst_bucket, new_name=folder_name)
            
            new_blob.acl.save(blob.acl)
            print(f"<{folder_name}> folder created.")

        except NotFound:
            print(f"Bucket <{bucket_name}> does not exist.")
    
    def delete_folder_from_bucket(self, bucket_name, folder_name):
        """This function is to create new folder from the bucket."""
        try:
            bucket = self.get_bucket(bucket_name)
            blobs = bucket.list_blobs(prefix=folder_name)
            for blob in blobs:
                blob.delete()
            
            print(f"Folder <{folder_name}> deleted.")

        except NotFound:
            print(f"Bucket <{bucket_name}> does not exist.")
    
    def delete_file_from_bucket(self, bucket_name, blob_name):
        """This function is to delete file from the bucket."""

        try:
            bucket = self.bucket(bucket_name)
            blob = bucket.blob(blob_name)
            blob.delete()
            print(f"Blob <{blob_name}> deleted.")

        except NotFound:
            print(f"No such file in this path: <{bucket_name}/{blob_name}>")


    @property
    def gcp_key(self):
        """This function is to get GCP key."""
        return self._gcp_key

    @gcp_key.setter
    def gcp_key(self, provided_gcp_key):
        """This function is to set GCP key."""
        if provided_gcp_key is not None:
            self._gcp_key = str(provided_gcp_key)
            self.credentials = str(provided_gcp_key)
        elif os.environ.get("SERVICE_ACCOUNT_KEY_PATH"):
            self._gcp_key = os.environ.get("SERVICE_ACCOUNT_KEY_PATH")
        else:
            self._gcp_key = None
            print(
                Fore.RED + "Warning!! GCP Key Path for Service Account is not specified"
            )

    @property
    def credentials(self):
        """This function is to get credentials."""
        return self._credentials

    @credentials.setter
    def credentials(self, provided_credentials):
        """This function is to set credentials."""
        if provided_credentials is not None:
            self._credentials = service_account.Credentials.from_service_account_file(
                self.gcp_key,
                scopes=["https://www.googleapis.com/auth/cloud-platform"],
            )
        elif os.environ.get("SERVICE_ACCOUNT_KEY_PATH"):
            self._credentials = service_account.Credentials.from_service_account_file(
                self.gcp_key,
                scopes=["https://www.googleapis.com/auth/cloud-platform"],
            )
        else:
            self._credentials = None
            print(
                Fore.RED + "Warning!! Credentials for Service Account is not specified"
            )
